#!/usr/bin/python
# -*- coding: UTF-8 -*-

## @package MainWindow.py
#  Paquete que implementa la clase de la ventana principal.
#  Incluye todos los mens, acciones y barras de herramientas, 
#  adems de los mdulos de edicin de texto como objetos hijos
#  de sta.
#
#  Para Toccata, IDE para Lilypond,
#  Futuro IDE para edicin completa de
#  Partituras, casi como WYSIWYG (espero)
#
#  @author Javier Salazar Loyola <jsalazar@elo.utfsm.cl>
############################################################

import sys

from Highlighter import *
from Subprocesos import *
from PyQt4 import QtGui, QtCore

## Clase heredera de QtGui.QMainWindow que contiene todos los mens,
#  editores y demases que hacen funcionar el programa, la instancia
#  de sta clase es la instancia madre de todos los dems objetos.
class MainWindow(QtGui.QMainWindow):

    ##Mtodo constructor
    # @param parent objeto padre
    # @param fileName Archivo que se cargar al momento de abrir el programa
    def __init__(self, parent=None, fileName = QtCore.QString('')):
        QtGui.QMainWindow.__init__(self,parent)

        ##Ttulo de la ventana. Para efectos de actualizacin del dem
        self.titulo = QtCore.QString('Toccata - %1')
        ##Nombre del archivo abierto en el momento
        self.archivoActual = QtCore.QString('')
                
        self.setWindowTitle('Toccata')
        self.setWindowIcon(QtGui.QIcon('images/Icono.png'))
        self.center()
        self.readSettings()

        self.statusBar()

        ##Bloque editor de texto. Encargado de mostrar el archivo .ly abierto
        self.textEdit = QtGui.QTextEdit()
        self.textEdit.setTabStopWidth(12)

        
        ##Instancia del resaltador de texto. El objeto que se encarga de mostrarlo
        # todo en colores
        # @see Highlighter.py
        self.textHighlighter = Highlighter()
        self.textHighlighter.addToDocument(self.textEdit.document())

        self.setCentralWidget(self.textEdit)

        self.crearDock()
        ##Objeto encargado de manejar los subprocesos requeridos. Por ahora, slo
        # estn soportado Lilypond, evince y xpdf (futuras versiones sern configurables)
        self.procesosHijos = ProcessHandler(parent = self)

        self.crearAcciones()
        self.crearMenuBar()
        self.crearToolBar()

        self.setupEditor()

        if not fileName.isEmpty():
            self.abrirArchivo(fileName)

    ##Mtodo que se encarga de cargar las configuraciones del programa.
    # Por ahora, est soportado slamente el tamao de la ventana y si es que al
    # momento de cerrarse (ultima configuracin guardada) estaba maximizada o no, adems
    # del ltimo directorio de trabajo
    def readSettings(self):
        settings = QtCore.QSettings("The O.C.I.O. Company Ltd","Toccata")
        size = settings.value("mainWindow/size",
                              QtCore.QVariant( QtCore.QSize(640, 480) )).toSize()
        isMax = settings.value("mainWindow/isMaximized",
                               QtCore.QVariant(False)).toBool()

        ##Directorio de trabajo. El directorio que se abrir por defecto con los cuadros
        # de dilogo y el directorio de trabajo de los procesos hijos
        self.directorio = QtCore.QDir(settings.value("Program/currentDir").toString())

        if isMax:
            self.showMaximized()
        else:
            self.resize(size)
        

    ##Mtodo que escribe la configuracin del programa. Escribe el tamao de la pantalla, si
    # est maximizada y el directorio de trabajo actual
    def writeSettings(self):
        settings = QtCore.QSettings("The O.C.I.O. Company Ltd", "Toccata")
        if not self.isMaximized():
            settings.setValue("mainWindow/size",QtCore.QVariant(self.size()))
            settings.setValue("mainWindow/isMaximized",QtCore.QVariant(False))
        else:
            settings.setValue("mainWindow/isMaximized", QtCore.QVariant(self.isMaximized()))
        settings.setValue("Program/currentDir",QtCore.QVariant(self.directorio.absolutePath()))
                

    ##Mtodo que configura el editor de texto y agrega los patrones y configuraciones de
    # color al objeto resaltador
    def setupEditor(self):
        #Funciones del programa
        funciones = QtGui.QTextCharFormat()
        funciones.setForeground(QtCore.Qt.red)
        self.textHighlighter.addMapping("\\\\[a-zA-Z!\(\)<>]+",funciones)
 
        #comentarios
        comentarios = QtGui.QTextCharFormat()
        comentarios.setFontItalic(True)
        comentarios.setForeground(QtCore.Qt.darkGreen)
        self.textHighlighter.addMapping("%.*[^\n]*",comentarios)

        #las notas... c d e f g a b
        notas = QtGui.QTextCharFormat()
        notas.setFontWeight(QtGui.QFont.Bold)
        notas.setForeground(QtCore.Qt.black)
        self.textHighlighter.addMapping("\\b[a-gr](es|is){0,2}[,']?\\d*\\b",notas)

        #comillas locas... texto y varios
        comillas = QtGui.QTextCharFormat()
        comillas.setFontItalic(True)
        comillas.setForeground(QtCore.Qt.darkRed)
        self.textHighlighter.addMapping("\".*\"",comillas)

        #smbolos raros de lilypond
        #se incluyen: {} [] , ' . - ! ? ^ / : = < > ( ) _
        simbolos = QtGui.QTextCharFormat()
        simbolos.setFontWeight(QtGui.QFont.Bold)
        simbolos.setForeground(QtCore.Qt.blue)
        self.textHighlighter.addMapping("[\{\}',\^\[\]!?\-\./:=<>\(\)_]", simbolos)
    
    ##Mtodo encargado de crear los mens y ordenarlos
    def crearMenuBar(self):
        ##La barra de men
        self.menubar = self.menuBar()

        ##Men archivo
        self.archivo = self.menubar.addMenu('&Archivo')
        self.archivo.addAction(self.nuevo)
        self.archivo.addAction(self.abrir)
        self.archivo.addSeparator()
        self.archivo.addAction(self.guardar)
        self.archivo.addAction(self.guardarComo)
        self.archivo.addSeparator()
        self.archivo.addAction(self.salir)

        ##Men edicin
        self.edicion = self.menubar.addMenu('&Edicion')
        self.edicion.addAction(self.deshacer)
        self.edicion.addAction(self.rehacer)
        self.edicion.addSeparator()
        self.edicion.addAction(self.copiar)
        self.edicion.addAction(self.cortar)
        self.edicion.addAction(self.pegar)
        self.edicion.addSeparator()
        self.edicion.addAction(self.selAll)

        ##Men compilacin
        self.compilacion = self.menubar.addMenu('&Compilacion')
        self.compilacion.addAction(self.lilypond)
        self.compilacion.addSeparator()
        self.compilacion.addAction(self.verPdf)
        self.compilacion.addAction(self.verPs)
        self.compilacion.addSeparator()
        self.compilacion.addAction(self.verLog)
        self.compilacion.addAction(self.saveLog)

        ##Men ayuda
        self.ayuda = self.menubar.addMenu('A&yuda')
        self.ayuda.addAction(self.acercaDe)
        self.ayuda.addAction(self.acercaDeQt)

    ##Mtodo encargado de crear la barra de herramientas a utilizar
    def crearToolBar(self):
        ##La barra de herramientas
        self.toolbar = self.addToolBar('Generico')

        self.toolbar.addAction(self.nuevo)
        self.toolbar.addAction(self.abrir)
        self.toolbar.addAction(self.guardar)
        self.toolbar.addSeparator()

        self.toolbar.addAction(self.deshacer)
        self.toolbar.addAction(self.rehacer)
        self.toolbar.addSeparator()

        self.toolbar.addAction(self.copiar)
        self.toolbar.addAction(self.cortar)
        self.toolbar.addAction(self.pegar)
        self.toolbar.addSeparator()

        self.toolbar.addAction(self.lilypond)
        self.toolbar.addAction(self.verPdf)
        self.toolbar.addSeparator()
        
        self.toolbar.addAction(self.salir)
        self.toolbar.addSeparator()
        self.toolbar.addAction(self.acercaDe)

    ##Mtodo que centra en la pantalla la ventana
    def center(self):
        screen = QtGui.QDesktopWidget().screenGeometry()
        size = self.geometry()
        self.move( (screen.width() - size.width())/2,
                   (screen.height() - size.height())/2)

    ##Mtodo que crea las acciones del programa y las conecta con sus correspondientes
    # mtodos
    def crearAcciones(self):
        #Men archivo
        ##Accin para salir del programa
        self.salir = QtGui.QAction(QtGui.QIcon('images/salir2.png'),'&Salir',
                                   self)
        self.salir.setShortcut('Ctrl+Q')
        self.salir.setStatusTip('Salir de la aplicacin')
        self.connect(self.salir, QtCore.SIGNAL('triggered()'),
                     QtCore.SLOT('close()'))

        ##Accin para guardar el archivo actual (previamente guardado o no)
        self.guardar = QtGui.QAction(QtGui.QIcon('images/save.png'),'&Guardar',
                                     self)
        self.guardar.setShortcut('Ctrl+S')
        self.guardar.setStatusTip('Guarda el archivo actual')
        self.connect(self.guardar, QtCore.SIGNAL('triggered()'),self.eventoGuardar)

        ##Accin para guardar el archivo actual eligiendo su nombre
        self.guardarComo = QtGui.QAction(QtGui.QIcon('images/save2.png'),'Guardar &como...',
                                         self)
        self.guardarComo.setShortcut('Ctrl+Alt+S')
        self.guardarComo.setStatusTip('Guarda el archivo eligiendo qu nombre tendr')
        self.connect(self.guardarComo, QtCore.SIGNAL('triggered()'),self.eventoGuardarComo)

        ##Accin para abrir un archivo
        self.abrir = QtGui.QAction(QtGui.QIcon('images/open.png'),'&Abrir',
                                   self)
        self.abrir.setShortcut('Ctrl+O')
        self.abrir.setStatusTip('Abre un archivo existente')
        self.connect(self.abrir,QtCore.SIGNAL('triggered()'),self.eventoAbrir)

        ##Accin para crear un archivo en blanco
        self.nuevo = QtGui.QAction(QtGui.QIcon('images/new.png'),'&Nuevo',
                                   self)
        self.nuevo.setShortcut('Ctrl+N')
        self.nuevo.setStatusTip('Crea un nuevo archivo')
        self.connect(self.nuevo, QtCore.SIGNAL('triggered()'),self.eventoNuevoArchivo)

        #Men Edicin
        ##Accin para deshacer la ltima edicin
        self.deshacer = QtGui.QAction(QtGui.QIcon('images/undo.png'), 'Des&hacer',self)
        self.deshacer.setShortcut('Ctrl+Z')
        self.deshacer.setStatusTip('Deshace la ltima edicin')
        self.connect(self.deshacer, QtCore.SIGNAL('triggered()'), self.undo)

        ##Accin para rehacer la ltima accin deshechada
        self.rehacer = QtGui.QAction(QtGui.QIcon('images/redo.png'),'&Rehacer',self)
        self.rehacer.setShortcut('Ctrl+Y')
        self.rehacer.setStatusTip('Rehace lo ltimo en haber sido deshecho')
        self.connect(self.rehacer, QtCore.SIGNAL('triggered()'), self.redo)
        
        ##Accin para cortar el texto seleccionado
        self.cortar = QtGui.QAction(QtGui.QIcon('images/cut.png'),'Cor&tar',self)
        self.cortar.setShortcut('Ctrl+X')
        self.cortar.setStatusTip('Corta el texto seleccionado para pegarlo con Pegar')
        self.connect(self.cortar, QtCore.SIGNAL('triggered()'),self.textEdit.cut)

        ##Accin para copiar el texto seleccionado
        self.copiar = QtGui.QAction(QtGui.QIcon('images/copy.png'), '&Copiar',self)
        self.copiar.setShortcut('Ctrl+C')
        self.copiar.setStatusTip('Copia el texto seleccionado para pegarlo con Pegar')
        self.connect(self.copiar, QtCore.SIGNAL('triggered()'),self.textEdit.copy)

        ##Accin para pegar el texto previamente copiado
        self.pegar = QtGui.QAction(QtGui.QIcon('images/paste.png'),'&Pegar', self)
        self.pegar.setShortcut('Ctrl+V')
        self.pegar.setStatusTip('Inserta un texto previamente copiado o cortado')
        self.connect(self.pegar, QtCore.SIGNAL('triggered()'),self.textEdit.paste)

        ##Accin para seleccionar todo el texto en el editor.
        self.selAll = QtGui.QAction('&Seleccionar Todo',self)
        self.selAll.setShortcut('Ctrl+A')
        self.selAll.setStatusTip('Selecciona todo el texto')
        self.connect(self.selAll, QtCore.SIGNAL('triggered()'),self.textEdit.selectAll)

        #detalles de algunos botones
        self.deshacer.setEnabled(False)
        self.rehacer.setEnabled(False)
        self.cortar.setEnabled(False)
        self.copiar.setEnabled(False)
        
        self.connect(self.textEdit, QtCore.SIGNAL("undoAvailable(bool)"),
                     self.deshacer.setEnabled)
        self.connect(self.textEdit, QtCore.SIGNAL("redoAvailable(bool)"),
                     self.rehacer.setEnabled)
        self.connect(self.textEdit, QtCore.SIGNAL("copyAvailable(bool)"),
                     self.cortar.setEnabled)
        self.connect(self.textEdit, QtCore.SIGNAL("copyAvailable(bool)"),
                     self.copiar.setEnabled)        

        #Men compilacin y debugging >.<
        ##Accin que inicia la compilacin del archivo actual
        self.lilypond = QtGui.QAction(QtGui.QIcon('images/icono2.png'),'&Lilypond',self)
        self.lilypond.setShortcut('F1')
        self.lilypond.setStatusTip('Compila el archivo actual')
        self.connect(self.lilypond, QtCore.SIGNAL('triggered()'),self.procesosHijos.compilar)

        ##Accin para el pdf creado
        self.verPdf = QtGui.QAction(QtGui.QIcon(''),'&Ver PDF', self)
        self.verPdf.setShortcut('F2')
        self.verPdf.setStatusTip('Cargar el PDF generado')
        self.connect(self.verPdf, QtCore.SIGNAL('triggered()'),self.procesosHijos.verPDF)

        ##Accin para el ps creado
        self.verPs = QtGui.QAction(QtGui.QIcon(''),'Ver &PS', self)
        self.verPs.setShortcut('F3')
        self.verPs.setStatusTip('Cargar el PS generado')
        self.connect(self.verPs, QtCore.SIGNAL('triggered()'),self.procesosHijos.verPS)

        ##Accin para mostrar el registro
        self.verLog = QtGui.QAction(QtGui.QIcon('images/new2.png'),'Ver &Registro', self)
        self.verLog.setShortcut('Ctrl+R')
        self.verLog.setStatusTip('Mostrar la salida de la compilacin con Lilypond')
        self.connect(self.verLog, QtCore.SIGNAL("triggered()"),self.mostrarLog)

        ##Accin para guardar el registro de compilacin
        self.saveLog = QtGui.QAction('&Guardar Registro',self)
        self.saveLog.setStatusTip('Guardar el resultado de la compilacin')
        self.connect(self.saveLog, QtCore.SIGNAL('triggered()'),self.LogView.salvarLog)
        

        #Men Ayuda
        ##Informacin acerca del programa
        self.acercaDe = QtGui.QAction(QtGui.QIcon('images/Icono.png'),
                                      '&Acerca de...',self)
        self.acercaDe.setShortcut('Ctrl+H')
        self.acercaDe.setStatusTip('Ayuda acerca de Toccata')
        self.connect(self.acercaDe, QtCore.SIGNAL('triggered()'),self.about)

        ##Informacin acerca de Qt
        self.acercaDeQt = QtGui.QAction(QtGui.QIcon('images/qt.png'),
                                        'Acerca de &Qt', self)
        self.acercaDeQt.setStatusTip('Informacin acerca de la herramienta Qt')
        self.connect(self.acercaDeQt, QtCore.SIGNAL('triggered()'),
                     QtGui.qApp.aboutQt)

    #Acciones que realiza la ventana esta
    ##Mtodo para deshacer
    def undo(self):
        doc = self.textEdit.document()
        doc.undo()
    ##Mtodo para rehacer
    def redo(self):
        doc = self.textEdit.document()
        doc.redo()
        
    ##Crea un archivo en blanco
    def eventoNuevoArchivo(self):
        #revisa que no haya nada escrito
        if self.textEdit.toPlainText().isEmpty():
            self.textEdit.clear()
            return
        
        else: #si hay, pregunta
            rpta = QtGui.QMessageBox.warning(self,'Nuevo', 'Guardar el archivo actual?',
                                             QtGui.QMessageBox.Yes,
                                             QtGui.QMessageBox.No, 
                                             QtGui.QMessageBox.Cancel | QtGui.QMessageBox.Default)
            if rpta == QtGui.QMessageBox.Yes:
                self.eventoGuardar()
            if rpta == QtGui.QMessageBox.No:
                pass
            if rpta == QtGui.QMessageBox.Cancel:
                return False
            self.textEdit.clear()
            self.archivoActual.clear()
            self.setWindowTitle('Toccata')
            return
        
        
    ##Abre un archivo existente.
    # @return False si no se seleccion archivo, en caso contrario, devuelve lo que devuelve
    # el mtodo abrirArchivo
    def eventoAbrir(self):
        fileName = QtGui.QFileDialog.getOpenFileName(self, 'Abrir...',
                                                     self.directorio.absolutePath(),
                                                     "Lylipond files (*.ly);;All Files (*)")
        if not fileName.isEmpty():
            return self.abrirArchivo(fileName)
        else:
            return False

    ##El mtodo que realiza la apertura efectiva del archivo.
    # @param fileName El nombre del archivo que se abrir
    # return True si se abre el archivo, False si no se puede
    def abrirArchivo(self,fileName):
        
        file = QtCore.QFile(fileName)
            
        #trata de abrir en modo de codificacin local y en modo rw
        if not file.open(QtCore.QFile.ReadWrite | QtCore.QFile.Text):
            QtGui.QMessageBox.warning(self,'Abrir',
                                      self.tr('No se puede abrir el archivo: %1').arg(fileName))
            return False

        #muestra cursor de ocupado despus de abrir las cosas
        #luego, lee todo el texto y restaura el cursor
        fileInfo = QtCore.QFileInfo(file)
        self.directorio = fileInfo.dir()

        texto = QtCore.QTextStream(file)
        QtGui.QApplication.setOverrideCursor(QtCore.Qt.WaitCursor)
        self.textEdit.setPlainText(texto.readAll())

        self.textHighlighter.loadHighlight(self.textEdit.document())
        
        QtGui.QApplication.restoreOverrideCursor()
        self.setCurrentFile(fileName)
        self.statusBar().showMessage('Archivo cargado exitosamente')
        return True

    ##Mtodo que pregunta el nombre del archivo a guardar y luego realiza el salvado del
    # archivo.
    # @return False si no se elige nombre para guardar, en otro caso, el valor de retorno de
    # eventoGuardarArchivo
    def eventoGuardarComo(self):
        #abre el cuadro de dilogo
        fileName = QtGui.QFileDialog.getSaveFileName(self,'Guardar como...',
                                                     self.directorio.absolutePath(),
                                                     "Lilypond Files (*.ly)")
        
        if fileName.isEmpty():
            return False
        else:
            return self.eventoGuardarArchivo(fileName)#guarda el archivo
            

    ##Mtodo que realiza el guardado del archivo. Si no se especifica, se aade automticamente
    # la extensin .ly al nombre del archivo.
    # @param fileName El nombre del archivo a guardar
    # @return True si se guarda, False en otro caso
    def eventoGuardarArchivo(self,fileName):
        if len(fileName) < 3 or fileName[-3:] != '.ly':#anexa el tipo de archivo por defecto
            fileName = fileName + '.ly'
        file = QtCore.QFile(fileName)

        if not file.open(QtCore.QFile.WriteOnly | QtCore.QFile.Text):#revisa que pueda guardarse
            QtGui.QMessageBox.critical(self, 'Guardar Archivo',
                                       'No se puede guardar el archivo en la ubicacin especificada')
            return False

        fileInfo = QtCore.QFileInfo(file)#salva el directorio actual
        self.directorio = fileInfo.dir()


        salida = QtCore.QTextStream(file)
        QtGui.QApplication.setOverrideCursor(QtCore.Qt.WaitCursor) #cursor de espera
        salida << self.textEdit.toPlainText()
        QtGui.QApplication.restoreOverrideCursor()

        self.setCurrentFile(fileName)
        self.statusBar().showMessage(self.tr('Archivo guardado: %1').arg(fileName))
        return True

    ##Mtodo llamado con la accin guardar: verifica que el archivo tenga o no nombre.
    # @return lo que devuelve el eventoGuardarComo si es que no hay archivo abierto, o
    # lo que devuelve eventoGuardarArchivo si es que hay ya un archivo abierto
    def eventoGuardar(self):
        if self.archivoActual.isEmpty():
            return self.eventoGuardarComo()
        else:
            return self.eventoGuardarArchivo(self.archivoActual)

    ##Coloca el ttulo de la ventana y actualiza el archivo actual.
    # @param fileName El nombre del archivo que se colocar en el ttulo
    def setCurrentFile(self,fileName):
        self.archivoActual = fileName
        if self.archivoActual.isEmpty():
            self.setWindowTitle('Toccata')
        else:
            self.setWindowTitle(self.titulo.arg(self.archivoActual))


    ##Mtodo que crea una subventana que muestra el registro de salida de la compilacin
    def crearDock(self):
        ##Subventana que contiene el visor de registro de salida
        self.dock = QtGui.QDockWidget("Registro de salida", self)
        self.dock.setAllowedAreas(QtCore.Qt.BottomDockWidgetArea | QtCore.Qt.TopDockWidgetArea)
               
        self.dock.setFeatures(QtGui.QDockWidget.DockWidgetClosable or
                              QtGui.QDockWidget.DockWidgetMovable)

        ##El registro de salida
        self.LogView = LogViewer(self.dock, self)
        self.dock.setWidget(self.LogView)
        self.addDockWidget(QtCore.Qt.BottomDockWidgetArea, self.dock)

    ##Mtodo que muestra de nuevo el registro
    def mostrarLog(self):
        self.dock.show()
        
    ##Mtodo que muestra informacin acerca de la aplicacin
    def about(self):
        QtGui.QMessageBox.about(self, 'Toccata', 'Primera versi&oacute;n de <b>Toccata</b>, GUI para editar partituras<br>Esta versi&oacute;n es una suerte de IDE para Lilypond<br>Por Javier Salazar Loyola')


    ##Define un evento de cierre diferente. Pregunta si sale en caso de haber modificado el
    # documento actual.
    # El objetivo es dejarlo como para que si hay cambios sin salvar,
    # que pregunte
    # @param event el evento de cierre que se llama
    # @return False si es que no se acepta el evento de cierre; en otro caso, no retorna nada
    def closeEvent(self,event):
        if self.textEdit.document().isModified():
            reply = QtGui.QMessageBox.question(self,'Pregunta...',
                                               "El documento ha sido modificado, guardar antes de salir?",
                                               QtGui.QMessageBox.Yes | QtGui.QMessageBox.Default,
                                               QtGui.QMessageBox.No,
                                               QtGui.QMessageBox.Cancel | QtGui.QMessageBox.Escape)

            if reply == QtGui.QMessageBox.Yes:
                if not self.eventoGuardar():
                    event.ignore
                    return False
                
                self.writeSettings()
                event.accept()
            else:
                if reply == QtGui.QMessageBox.No:
                    self.writeSettings()
                    event.accept()
                else:
                    event.ignore()
                    return
                return
            return
        else:
            self.writeSettings()
            event.accept()


if __name__ == "__main__":
    app = QtGui.QApplication(sys.argv)
    if len(sys.argv) > 1:
        mainwindow = MainWindow(fileName = QtCore.QString(sys.argv[1]))
    else:
        mainwindow = MainWindow()
    mainwindow.show()
    sys.exit(app.exec_())
